/* collector.c - collector main
 *
 * Copyright (C) 2006-2007 Abilis
 * Copyright (C) 2006-2007 Antek srl
 *
 * Author: Elisa Manara <e@entropika.net>
 * Last update: Mon Dec 11 10:01:26 CET 2006
 *
 * This file is part of Abilis Collector for *nix.
 *
 * Abilis Collector is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version.
 *
 * Abilis Collector is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
 * See the GNU General Public License for more details.
 *                                   
 * You should have received a copy of the GNU General Public License along                        
 * with Abilis Collector; see the file COPYING.  If not, write to the Free                        
 * Software Foundation, 59 Temple Place, Suite 330, Boston, MA 02111 USA. 
 */  

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "define.h"

int recvPacket(struct CfgClient *client, unsigned int *cmd, unsigned int *method,
               unsigned char *data, unsigned int maxdatalen);
int handlePacket(struct CfgData *config, struct CfgClient *client, 
    unsigned int cmd, unsigned int method, unsigned char *data, 
    unsigned int datalen); 

int mysqlStart(struct CfgData *config);
int mysqlStop(struct CfgData *config);
int queueInit(struct CfgData *config);
int queueProcess(struct CfgData *config, time_t now);

int collectorStart(int argc, char **argv, struct CfgData *config, 
    struct CfgClient **client);
int collectorHandleSignals(int argc, char **argv, struct CfgData *config, 
    struct CfgClient *client);
void collectorSignalsConnect(int *signals);

int connectToAgent(struct CfgClient *client, time_t now);
int sendAuthentication(struct CfgClient *client, char *userdata);
int checkAgentAuthentication(struct CfgClient *client, unsigned int cmd, unsigned int m,
    unsigned char *data, unsigned int len);
int sendModesRequest(struct CfgClient *client);
int checkSupportedModes(struct CfgClient *client, unsigned int cmd, unsigned int m,
    unsigned char *data, unsigned int len);
int setMode(struct CfgClient *client, int mode);
int checkModeAcceptance(struct CfgClient *client, unsigned int cmd, unsigned int m,
    unsigned char *data, unsigned int len);
int sendDummy(struct CfgClient *client);
void sessionKeepAlive(struct CfgClient *client, time_t now);
int sendTerminateSession(struct CfgClient *client, char *reason);
int checkAgentTermination(struct CfgClient *client, unsigned int cmd, unsigned int m,
    unsigned char *data, unsigned int len);

void collectorStatus(struct CfgData *config, struct CfgClient *client);
void collectorTerminate(struct CfgData *config, struct CfgClient *client);
void collectorEnd(struct CfgData *config, struct CfgClient *client);
int collectorSignalRequest(int argc, char **argv, pid_t c_pid);
int collectorSignalsBlock(int *signals);
int collectorSignalsUnblock(int *signals);
void baseName(char *path, char *basename);

static void collectorInit(struct CfgData *config);
static void fdSet(struct CfgData *config, struct CfgClient *client, fd_set *reads);
static int maxSocket(struct CfgData *config, struct CfgClient *client);
static int loadConfig(struct CfgData *config, struct CfgClient **client);
static pid_t collectorCheckPid(char *cmdline);
static int collectorWritePid(void);
static int collectorDeletePid(void);

/* global debug - override from configuration value */
int DebugFlag = 0;

/* client array */
static struct CfgClient *client = NULL;
/* default client values */
static struct CfgClient defaultClient = {
	.id = "",
	.agentHost = "",
	.agentPort = 0,
	.mysqlTable = "",
	.error = "",
	.status = -1,
	.fsm = DISCONNECTED,
	.reconnect = 0,
	.timeo = RETRY_DISABLED,
	.alive = 0,
};

static struct {
	int action;
	int err;
} shutdown;



void annihilate(void) {
	if (client != NULL) {
		printDebug("%s: free clients data\n", __func__);
		free(client);
		client = NULL;
	}
}

void setShutdown(int action, int retval) {
	shutdown.action = action;
	shutdown.err = retval;
}


int main(int argc, char **argv) {
	struct CfgData config;
	struct timeval tmout;
	int c, i, retval, len;
	unsigned int command, method;
	unsigned char data[PACKET_MAX_SIZE];
	char userdata[AUTH_DATA_SIZE] = "xxx";
	time_t now;
	fd_set reads;
	int signals[] = {SIGHUP, SIGINT, SIGQUIT,
	     SIGILL, SIGSEGV, SIGPIPE,
	     SIGTERM, SIGUSR1};
	int all_agents_closed;

	/* set signals */
	collectorSignalsConnect(signals);

	do {

		all_agents_closed = 0;
		/* initialize structs, load configuration, start MySQL link */
		retval = collectorStart(argc, argv, &config, &client);
		if (retval < 0) {
			printError("Cannot execute collector. Exiting now.\n");
			mysqlStop(&config);
			all_agents_closed = 1;
			setShutdown(COLLECTOR_HALT, 1);
		} else if (retval == 1) {
			mysqlStop(&config);
			all_agents_closed = 1;
			setShutdown(COLLECTOR_HALT, 0);
		}

		/* initialize query cache */
		queueInit(&config);

		while (!all_agents_closed) {
			/* signals handler */
			collectorHandleSignals(argc, argv, &config, client);

			/* current time */
			now = time(NULL);

			/* process query queue, if exists */
			queueProcess(&config, now);

			for (i=0; i<config.clients; i++) {

				if (client[i].timeo != RETRY_DISABLED && client[i].timeo < now) {
					printError("[%s] Agent %s not responding\n",
					    client[i].id, client[i].agentHost);
					setInactive(&client[i], "Agent not responding");
				}
				retval = -1;
				switch (client[i].fsm) {
				case TERMINATE:
					if (!isActive(&client[i])) {
						client[i].fsm = CLOSED;
						client[i].timeo = RETRY_DISABLED;
					} else {
						retval = sendTerminateSession(
						    &client[i], "Collector must quit");
						if (retval == 0) {
							client[i].fsm = TERMINATION_SENT;
							client[i].timeo = now + SESSION_END_TIMEOUT;
						}
					}
					break;
				case DISCONNECTED:
					retval = connectToAgent(&client[i], now);
					if (retval < 0) {
						client[i].reconnect = now + config.retry;
						client[i].timeo = RETRY_DISABLED;
					} else if (retval == 0) {
						client[i].fsm = CONNECTED;
						client[i].timeo = now + RETRY_MESSAGE;
					}
					break;
				case CONNECTED:
					retval = sendAuthentication(&client[i], userdata);
					if (retval == 0) {
						client[i].fsm = AUTHENTICATION_SENT;
						client[i].timeo = now + RETRY_MESSAGE;
					}
					break;
				case AUTHENTICATED_AGENT:
					retval = sendModesRequest(&client[i]);
					if (retval == 0) {
						client[i].fsm = MODES_SENT;
						client[i].timeo = now + RETRY_MESSAGE;
					}
					break;
				case MODE_NEGOTIATED:
					retval = setMode(&client[i], SIMPLE_MODE_2K);
					if (retval == 0) {
						client[i].fsm = MODE_SET;
						client[i].timeo = now + RETRY_MESSAGE;
					}
					break;
				case RUNNING:
					/* send dummy message as keep alive for client<=>agent session */
					sessionKeepAlive(&client[i], now);
					client[i].timeo = RETRY_DISABLED;
					break;
				case ALIVE_RECV:
					sendDummy(&client[i]);
					printDebug("[%s] Keep alive (response) sent to %s\n",
					    client[i].id, client[i].agentHost);
					client[i].timeo = RETRY_DISABLED;
					client[i].fsm = RUNNING;
					break;
				default:
					break;
				}

			}

			c = maxSocket(&config, client)+1;
			#if 0
			if(c == 1) {
			    printError("No active connection to agent(s). I'll retry in %i seconds.\n",
				config.retry);
			    sleep(config.retry);
			    continue;
			}
			#endif

			/* only connected && authenticated clients here */
			FD_ZERO(&reads);
			fdSet(&config, client, &reads);

			/* set timeout for select == 1 second */
			tmout.tv_sec = POLLING_TIMEO_SEC;
			tmout.tv_usec = 0;

			/* block all signals, thus they don't interfere with select() */
			collectorSignalsBlock(signals);
			retval=select(c, &reads, NULL, NULL, &tmout);
			/* unblock signals */
			collectorSignalsUnblock(signals);

			# if 0
			if (retval == 0)
				/* no socket ready for read */
				continue;
			#endif
			if (retval == -1) {
				printError("Select: %s", strerror(errno));
				/* TODO terminate or retry later? */
				collectorTerminate(&config, client);
				setShutdown(COLLECTOR_HALT, 1);
			}

			for (i=0; i<config.clients; i++) {
				if (!isActive(&client[i]) ||
				    !FD_ISSET(client[i].status, &reads))
					continue;
				len = recvPacket(&client[i], &command, &method, data, PACKET_MAX_SIZE);
				if (len == 0) {
					/* remote side closed connection */
					printError("[%s] Agent %s closed connection!\n", 
					    client[i].id, client[i].agentHost);
					setInactive(&client[i], "Remote side closed connection");
				}
				if (len == -1) {
          /* disconnect agent in order to force reconnection */
					printError("[%s] Error receiving packet\n", client[i].id);
					setInactive(&client[i], "Error receiving packet");
        }

				retval = -1;
				switch (client[i].fsm) {
				case TERMINATION_SENT:
					retval = checkAgentTermination(&client[i], command,
					    method, data, len);
					if (retval == 0) {
						client[i].fsm = CLOSED;
						client[i].timeo = RETRY_DISABLED;
					}
					break;
				case AUTHENTICATION_SENT:
					retval = checkAgentAuthentication(&client[i], command,
					    method, data, len);
					if (retval == 0) {
						client[i].fsm = AUTHENTICATED_AGENT;
						client[i].timeo = now + RETRY_MESSAGE;
					}
					break;
				case MODES_SENT:
					retval = checkSupportedModes(&client[i], command,
					    method, data, len);
					if (retval == 0) {
						client[i].fsm = MODE_NEGOTIATED;
						client[i].timeo = now + RETRY_MESSAGE;
					}
					break;
				case MODE_SET:
					retval = checkModeAcceptance(&client[i], command,
					    method, data, len);
					if (retval == 0) {
						printDebug("[%s] Connection to agent %s succeeds\n",
						    client[i].id, client[i].agentHost);
						client[i].fsm = RUNNING;
						client[i].timeo = RETRY_DISABLED;
					}
					break;
				case RUNNING:
					/* what kind of packet is this? ;) */
					handlePacket(&config, &client[i], command, method, data, len);
					client[i].timeo = RETRY_DISABLED;
					break;
				default:
					break;
				}
			}

			collectorSignalsBlock(signals);
			all_agents_closed = 1;
			for (i=0; i<config.clients; i++)
				if (client[i].fsm != CLOSED) {
					all_agents_closed = 0;
					break;
				}
			collectorSignalsUnblock(signals);

			if (all_agents_closed)
				collectorEnd(&config, client);

		}

		printDebug("All clients closed. Action = %i\n", shutdown.action);

		annihilate();
	} while (shutdown.action != COLLECTOR_HALT);
	exit(shutdown.err);
}

int collectorStart(int argc, char **argv, struct CfgData *config, 
    struct CfgClient **client)
{
    int retval;
    pid_t running_pid;

    /* initialize config struct */
    collectorInit(config);

    /* load configuration parameters */
    if(loadConfig(config, client) != 0) {
	printError("Cannot load configuration. Must exit.\n");
	return -1;
    }

    printDebug("Collector version: %s\n",COLLECTOR_VERSION);

    /* check collector instance */
    running_pid = collectorCheckPid(argv[0]);
    switch(running_pid) {
    case -1:
	printError("Pidfile corrupted?\n");
	return -1;
	break;
    case 0:
        /* no other collector is running!
	 * If we have a request from command line (i.e. status,
	 * reload, stop) exit.
	 */
	if(argc > 1) {
	    printError("Sorry, no collector is currently running.\n");
	    return -1;
	}
	/* ok, a new collector is starting (write a new pidfile) */
	if((retval=collectorWritePid()) != 0) {
	    printError("Cannot write pidfile. Must exit.\n");
	    return -1;
	}
	break;
    default:
	/* one collector istance is already running:
	 * we check if there is a good reason for calling it ;)
	 */
	retval = collectorSignalRequest(argc, argv, running_pid);
	mysqlStop(config);
	return (retval);
	break;
    }

    /* update start time */
    //strftime(config->uptime, 32, "%F %T", localtime(&now));
    getHumanTime(config->uptime);


    /* start mySQL connection */
    retval = mysqlStart(config);
    if(retval == -1) {
	printError("[MySQL] Connection failed to host %s.\n",
	    config->mysqlHost);
	collectorDeletePid();
	return -1;
    }
    return 0;
}

pid_t collectorCheckPid(char *cmdline)
{
    char cname[MAX_PATH_LEN];
    char pname[MAX_PATH_LEN];
    char procpath[MAX_PATH_LEN];
    char procline[MAX_PATH_LEN];
    FILE *p_file;
    pid_t pid;
    
    /* check if a pid_file exists, i.e. the program is already running */
    if((p_file=fopen(PID_FILE, "r")) == NULL) {
	if(errno != ENOENT) {
	    printError("Pid file %s: %s\n", PID_FILE, strerror(errno));
	    return -1;
	}
	/* errno == ENOENT: pidfile does not exist, 
	 * thus no collector is running
	 */
	return 0;
    }

    /* another collector is already running */
    /* read the pid from pidfile */
    fscanf(p_file, "%d\n", &pid);
    fclose(p_file);

    /* check pid format... */
    if(pid == 0 || pid == EOF) return -1;

/* only for system that use proc pseudo-filesystem */
#if USE_PROCFS
    /* proc pseudo-filesystem check */
    snprintf(procpath, MAX_PATH_LEN, "/proc/%i/cmdline", pid);
    if((p_file=fopen(procpath, "r")) == NULL) {
	if(errno != ENOENT) {
	    printError("Error reading %s: %s\n", procpath, strerror(errno));
	    return -1;
	}
	/* errno == ENOENT: process does not exist...*/
	collectorDeletePid();
	return 0;
    }

    /* read command line in proc */
    fscanf(p_file, "%s\n", procline);
    fclose(p_file);
    
    baseName(cmdline, cname);
    baseName(procline, pname);

    /* compare process names */
    if(strcasecmp(cname, pname)) {
	printError("Pid refers to a different process: %s\n", pname);
	collectorDeletePid();
	return 0;
    }
#endif

    return pid;
}

int collectorWritePid(void)
{
    pid_t pid;
    FILE *p_file;

    pid = getpid();

    if((p_file=fopen(PID_FILE, "w")) == NULL) {
	printError("Cannot open %s for writing: %s\n", 
	    PID_FILE, strerror(errno));
	return -1;
    }

    fprintf(p_file, "%d\n", pid);
    fclose(p_file);

    /* rw permission only to file owner */
    chmod(PID_FILE, S_IRUSR|S_IWUSR);
    printDebug("Wrote pid %i on pidfile %s\n", pid, PID_FILE);
    return 0;
}

int collectorDeletePid(void)
{
    int retval;

    if((retval = unlink(PID_FILE)) < 0) {
	printError("Error unlinking pid file %s: %s\n", 
	    PID_FILE, strerror(errno));
	return -1;
    }
    printDebug("Pidfile unlinked: %s\n", PID_FILE);
    return 0;
}

void collectorTerminate(struct CfgData *config, struct CfgClient *client) {
    int i;

    for(i = 0; i < config->clients; i++) {
	client[i].fsm = TERMINATE;
	client[i].timeo = time(NULL) + RETRY_MESSAGE;
    }
}

void collectorEnd(struct CfgData *config, struct CfgClient *client)
{
    int i;

    mysqlStop(config);
    printDebug("Closing client<=>agent connections...\n");
    for(i = 0; i < config->clients; i++) {
	if(isActive(&client[i])) {
	    setInactive(&client[i], "Resetted");
	}
    }
    /* delete pidfile */
    collectorDeletePid();

    printDebug("Done.\n");
}

/* Initialize configuration struct */
void collectorInit(struct CfgData *config)
{
    memset(config, 0, sizeof(*config));
    memset(&config->mysqlLink, 0, sizeof(config->mysqlLink));
    config->clients = 0;
}

/* Load configuration parameters from defined config file.
 * Return 0 on success, -1 otherwise
 */
int loadConfig(struct CfgData *config, struct CfgClient **client)
{
    FILE *fp;
    char buf[1024];
    int i, ret;
    int entry = -1;
    char config_file[MAX_PATH_LEN];
    char *fields[] = { "MYSQLHOST:", "MYSQLPORT:", "MYSQLUSER:", 
			"MYSQLPASS:", "MYSQLDB:", "AGENTHOST:", 
			"AGENTPORT:", "MYSQLTABLE:", "CLIENT:", 
			"RETRY:", "DEBUG:", NULL };
    #define CLIENT_FIELD	8

    snprintf(config_file, MAX_PATH_LEN, "%s", CONFIG_FILE);
    if((fp = fopen(config_file, "r")) == NULL) {
	printError("Error reading config file %s: %s\n",
	    CONFIG_FILE, strerror(errno));
	return -1;
    }

    while (fgets(buf, 1024, fp) != NULL) {
	char *line = buf;
	int l = strlen(buf);
	if (l && buf[l-1] == '\n')
	    buf[l-1] = '\0';

	if (strncmp(line, fields[CLIENT_FIELD], strlen(fields[CLIENT_FIELD])) != 0)
	    continue;

	config->clients++;
    }
    rewind(fp);
    *client = malloc(sizeof(struct CfgClient)*config->clients);
    if (*client == NULL) {
	printError("%s: %s\n", __func__, strerror(errno));
	return (-1);
    }
    for (i=0; i<config->clients; i++)
	memcpy(&(*client)[i], &defaultClient, sizeof(struct CfgClient));

    ret = 0;
    while((fgets(buf, 1024, fp)) != NULL)
    {
	char *line = buf;

	int l = strlen(buf);
	if (l && buf[l-1] == '\n')
	    buf[l-1] = '\0';

	for(i = 0; fields[i]; i++) {
	    if((strncmp(line, fields[i], strlen(fields[i]))) != 0)
		continue;

	    line+=strlen(fields[i]);
	    /* global section */
	    if ((strcmp(fields[i], "MYSQLHOST:")) == 0) {
		snprintf(config->mysqlHost,MAX_CONFVAL_LEN,"%s",line);
	    } else if ((strcmp(fields[i], "MYSQLPORT:")) == 0) {
		config->mysqlPort = atoi(line);
	    } else if ((strcmp(fields[i], "MYSQLUSER:")) == 0) {
		snprintf(config->mysqlUser,MAX_CONFVAL_LEN,"%s",line);
	    } else if ((strcmp(fields[i], "MYSQLPASS:")) == 0) {
		snprintf(config->mysqlPass,MAX_CONFVAL_LEN,"%s",line);
	    } else if ((strcmp(fields[i], "MYSQLDB:")) == 0) {
		snprintf(config->mysqlDb,MAX_CONFVAL_LEN,"%s",line);
	    } else if ((strcmp(fields[i], "DEBUG:")) == 0) {
		DebugFlag = atoi(line);
	    } else if ((strcmp(fields[i], "RETRY:")) == 0) {
		config->retry = atoi(line);
	    /* clients section */
	    } else if ((strcmp(fields[i], "CLIENT:")) == 0) {
		/* new client entry */
		entry++;
		/* max agent entry reached */
		if (entry>=config->clients) {
		    printError("%s: wrong estimated number of clients: %li\n",
			__func__, config->clients);
		    ret = -1;
		    break;
		}
		snprintf((*client)[entry].id,MAX_CLIENTID_LEN,"%s",line);
	    } else if((strcmp(fields[i], "AGENTHOST:")) == 0) {
		snprintf((*client)[entry].agentHost,MAX_CONFVAL_LEN,"%s",line);
	    } else if ((strcmp(fields[i], "AGENTPORT:")) == 0) {
		(*client)[entry].agentPort = atoi(line);
	    } else if ((strcmp(fields[i], "MYSQLTABLE:")) == 0) {
		snprintf((*client)[entry].mysqlTable,MAX_CONFVAL_LEN,"%s",line);
	    } else {
		printError("Unknown keyword '%s' in config file\n", line);
	    }
	}
	if (ret<0)
	    break;
    }
    fclose(fp);

    printDebug("Number of clients supported: %li\n", config->clients);
    if ((*client)[0].id[0] == '\0') {
	printError("No configured host\n");
	ret = -1;
    }
    return (ret);
}

/* dump client status and configuration, queue data.
 * (You can call collectorStatus() by sending a SIGUSR1
 * to running process; example:
 * kill -SIGUSR1 <process pid>
 */
void collectorStatus(struct CfgData *config, struct CfgClient *client)
{
    int i;

    printf("Collector version %s\n", COLLECTOR_VERSION);

    printf("\n***** Global section *****\n");
    printf("UPTIME     = %s\n", config->uptime);
    printf("MYSQLHOST  = %s\n", config->mysqlHost);
    printf("MYSQLPORT  = %i\n", config->mysqlPort);
    printf("MYSQLDB    = %s\n", config->mysqlDb);
    printf("MYSQLUSER  = %s\n", config->mysqlUser);
    printf("MYSQLPASS  = %s\n", config->mysqlPass);
    printf("RETRY      = %i\n", config->retry);
    printf("DEBUG FLAG = %i\n", DebugFlag);

    printf("\n***** Queue section *****\n");
    printf("Queued    = %i\n", config->queue.length);
    printf("Processed = %i\n", config->queue.processed);
    printf("Inserted  = %i\n", config->queue.inserted);
    printf("Duplicate = %i\n", config->queue.duplicate);
    printf("Skipped   = %i\n", config->queue.skipped);

    printf("\n***** Client section *****\n");
    for(i = 0; i < config->clients; i++) {
	if(client[i].id[0] == '\0')
	    break;
	printf("CLIENT[%i]  = %s\n", i, client[i].agentHost);
	printf("AGENTHOST[%i]  = %s\n", i, client[i].agentHost);
	printf("AGENTPORT[%i]  = %i\n", i, client[i].agentPort);
	printf("MYSQLTABLE[%i]  = %s\n", i, client[i].mysqlTable);
	printf("RECONNECTION [%i] = %i\n", i, client[i].reconnect);
	printf("FSM [%i] = %i\n", i, client[i].fsm);
	printf("TIMEO [%i] = %i\n", i, client[i].timeo);
	printf("STATUS FLAG[%i] = %i\n", i, client[i].status);
	printf("LAST ERROR[%i] = %s\n", i, client[i].error);
	printf("---\n");
    }
    if(fflush(stdout))
	printError("fflush: %s\n", strerror(errno));
}

int isActive(struct CfgClient *client)
{
    if(client->status > 0)
	return 1;
    else
	return 0;
}

int isConfigured(struct CfgClient *client)
{
    if(client->id[0] != '\0')
	return 1;
    else
	return 0;
}

void setError(struct CfgClient *client, char *error)
{
    snprintf(client->error, ERROR_LEN, "%s", error);
    return;
}

/* set as inactive a configuration client entry:
 * 1) close the socket;
 * 2) set status (socket) = -1;
 * 3) call setError() in order to set related error message for client
 */
void setInactive(struct CfgClient *client, char *error)
{
    if (client->status >= 0)
	close(client->status);
    client->status = -1;

    if (client->fsm == TERMINATE ||
	client->fsm == TERMINATION_SENT ||
	client->fsm == CLOSED) {
	    client->fsm = CLOSED;
	    client->timeo = RETRY_DISABLED;
    } else {
	client->fsm = DISCONNECTED;
	client->timeo = RETRY_DISABLED;
    }

    setError(client, error);
}

int maxSocket(struct CfgData *config, struct CfgClient *client)
{
    int i;
    int max = 0;
    for(i = 0; i < config->clients; i++) {
	if(client[i].status > max)
	    max = client[i].status;
    }
    return max;
}

void fdSet(struct CfgData *config, struct CfgClient *client, fd_set *reads)
{
    int i;
    for(i = 0; i < config->clients; i++) {
	if(isActive(&client[i]))
	    FD_SET(client[i].status, reads);
    }
}

/* 
 * vim: softtabstop=4 shiftwidth=4 
 */


